def ordena_iniciativa(lista: dict[str:int], maior_ganha: bool = True) -> list[str]:
    """
    :param lista: dicionário com nome do personagem e o valor rolado
    :param maior_ganha: True por padrão, determina se quem vence a iniciativa é o maior valor
    :return: lista dos nomes ordenadas por quem vence a iniciativa
    """
    lista = sorted(lista.items(), key=lambda x: x[1], reverse=maior_ganha)
    return [personagem[0] for personagem in lista]


if __name__ == '__main__':
    from dado.dados import d20

    iniciativa = {
        'PC 1': d20() + 2.5,
        'PC 2': d20() + 1,
        'PC 3': d20() + 3,
        'NPC 1': d20() - 1,
        'NPC 2': d20() + 5
    }

    print(iniciativa)
    print(ordena_iniciativa(iniciativa))
    print(ordena_iniciativa({'nome': 2, 'nome 2': 10}))



from dataclasses import dataclass


@dataclass
class DadoRolado:
    """
    Retorna uma dataclass composto pelo rolamento, o dado rolado e seu valor e o valor total
    """
    rolamento: int
    modificador: int = 0
    total: int = 0

    def __post_init__(self):
        self.total = self.rolamento + self.modificador



from RPG.dado.rolar_dado_notacao import rolar_dado_notacao
import re


def verifica_resultado_fora_alcance(valor_rolado, indice_menor, indice_maior, tabela, resultado):
    """
    Caso o rolamento seja fora do alcance pega o menor valor se estiver abaixo e o maior valor se estiver acima

    :param valor_rolado:
    :param indice_menor:
    :param indice_maior:
    :param tabela:
    :param resultado:
    :return:
    """
    if not resultado and (valor_rolado < indice_menor or valor_rolado > indice_maior):
        indice = 0 if valor_rolado < indice_menor else -1
        return tabela[indice][1]
    return resultado


def rola_dados_resultado(rolar_dados: bool, resultado: str) -> str:
    """
    Transforma o texto que contenha notação de _dados em texto com resultado
    :param rolar_dados:
    :param resultado:
    :return:
    """
    if rolar_dados and type(resultado) == str:
        palavras = resultado.split()
        texto_regex = []

        for palavra in palavras:
            if re.match(r'\d*d\d+(?:[-+*/]\d+)?', palavra):
                rolado = rolar_dado_notacao(palavra)
                texto_regex.append(str(rolado))
            else:
                texto_regex.append(palavra)

        resultado = ' '.join(texto_regex)

    return resultado



from dataclasses import dataclass
from RPG import rolar_dado_notacao, TabelaRolavel

from .data import tesouro, mecanicas


@dataclass
class Tesouro:
    _cr: int = 1

    def __post_init__(self):
        self._inicia_param()

    def _inicia_param(self):
        self.moedas = {'cp': 0, 'sp': 0, 'ep': 0, 'gp': 0, 'pp': 0}
        self.bens = None
        self.itens = None

    def _rola_moedas(self, tabela):
        moedas = [rolar_dado_notacao(moeda) if type(moeda) == str else moeda for moeda in tabela]

        for i in range(len(moedas)):
            self.moedas[mecanicas.tipos_moedas[i]] += moedas[i]


@dataclass
class TesouroIndividual(Tesouro):
    def __post_init__(self):
        super().__post_init__()
        self._rolar()

    def _rolar(self):
        self._inicia_param()

        tabela_individual = TabelaRolavel(tesouro.individual, self._cr).resultado
        tabela_moedas = TabelaRolavel(tabela_individual, 'd100').resultado

        self._rola_moedas(tabela_moedas)


@dataclass
class TesouroHoard(Tesouro):
    pass


import random


class Dado:
    def __init__(self, quantidade: int = 1, faces: int = 6):
        """
        Gera uma lista dos _dados rolados

        rolamento = Dado(quantidade=2, faces=6)

        print(rolamento) -> "[3, 5]"  # valor aleatório

        Caso use algum método para alterar o resultado da lista, é possível acessar
        com o atributo '.rolamento_original'

        :param quantidade: a quantidade de _dados rolados, padrão: 1
        :param faces: o número de faces do dado, padrão: 6
        """
        self._quantidade = quantidade
        self._faces = faces
        self._rolamento = self._rolagem_de_dados()
        self.rolamento_original = tuple(self._rolamento)

    def __str__(self):
        return f'{self._quantidade}d{self._faces}, _dados: {self._rolamento}, total: {self.total}'

    def __repr__(self):
        return f'<Dado quantidade={self._quantidade} faces={self._faces} rolamento={self._rolamento}>'

    def __len__(self):
        return len(self._rolamento)

    def __getitem__(self, index):
        return self._rolamento[index]

    def __setitem__(self, index, value):
        self._rolamento[index] = value

    def __lt__(self, other):
        return self.total < other.total

    # Funções de abertura ======================================================
    def _rolagem_de_dados(self):
        """
        Retorna uma lista com os _dados rolados
        """
        return [random.randint(1, self._faces) for _ in range(self._quantidade)]

    # Métodos para alterar o resultado =========================================
    def rerolar(self):
        """
        Rerola os dados, com a quantidade e face originais
        """
        self._rolamento.clear()
        self._rolamento.extend(self._rolagem_de_dados())

    def retirar_menor(self):
        """
        Retira o menor dado rolado da lista
        """
        self._rolamento.remove(min(self._rolamento))

    def explodir(self, explode_em: int = None):
        """
        :param explode_em: intenger, explode acima ou igual a esse valor. Por padrão é o número de faces do dado
        """
        if explode_em is None:
            explode_em = self._faces

        for i in self._rolamento:
            if i >= explode_em:
                self._rolamento.append(random.randint(1, self._faces))

    # getters ==================================================================
    @property
    def media(self):
        base = (1 + self._faces) / 2
        return int(base * self._quantidade)

    @property
    def total(self):
        return sum(self._rolamento)

    @property
    def maior(self):
        return max(self._rolamento)

    @property
    def menor(self):
        return min(self._rolamento)


if __name__ == "__main__":
    # Exemplo de uso
    print(Dado(5, 6))
    print("A média de 5d6 é:", Dado(5, 6).media)

    print("------------------------------------------------------------")
    print(f"Rolando 3d4, mas explosivo:")
    print(Dado(3, 4).explodir())

    print("------------------------------------------------------------")
    print("Rolando 4d6, 6 vezes e tirando o menor dado:")
    atributos_dnd = {Dado(4, 6) for _ in range(6)}
    atributos_final = []

    for dado in atributos_dnd:
        dado.retirar_menor()
        atributos_final.append(dado.total)

    print(atributos_final)



from .dado import Dado
from helpers.converter_notacao_dado import converter_notacao_dado


# ==================================================================
# Funções relacionadas
# ==================================================================

def rolar_dado_notacao(notacao: str) -> int:
    """
    Rola um conjunto de _dados com base na linguagem de RPG, estilo "3d6+3" e retorna o valor total

    :param notacao: A fórmula do dado. Aceita mais de um dado, como "1d8+1d6"
    """
    # converte a notação em _dados
    notacao = converter_notacao_dado(notacao)

    resultado = 0

    # insere os _dados
    for dado in notacao.get("_dados"):
        quantidade = dado.get("quantidade")
        faces = dado.get("faces")
        operador = dado.get("operador") if dado.get("operador") else "+"

        total_rolado = Dado(quantidade, faces).total

        resultado = eval(f'{resultado} {operador} {total_rolado}')

    # insere o modificador_dado
    operador_modificador = notacao.get("modificador_dado").get("operador")
    valor_modificador = notacao.get("modificador_dado").get("valor")

    resultado = eval(f'{resultado} {operador_modificador} {valor_modificador}')

    return resultado


if __name__ == '__main__':
    print(rolar_dado_notacao('3d6'))



import random


class Baralho:
    def __init__(self, cartas: list):
        """
        Manipula uma lista de cartas_aplicadas, embaralhando seus elementos e permiti sacar, resetar para seu status original

        :param cartas: a lista de cartas_aplicadas a ser usada
        """

        self._cartas = cartas
        self._original = tuple(cartas)

        self.descarte = list()
        self.em_jogo = list()

        self.reiniciar_baralho()

    def __str__(self):
        return f'{len(self._cartas)} cartas disponíveis para saque: {self._cartas}'

    def __repr__(self):
        return f'<Baralho com as cartas {self._original}>'

    def __len__(self):
        return len(self._cartas)

    def __getitem__(self, item):
        return self._cartas[item]

    # GETTERS ==============================================================
    @property
    def pilha_saque(self):
        return self._cartas

    # MÉTODOS DO BARALHO ===================================================
    # Esses métodos são relacionados somente com o baralho e não cartas
    def reiniciar_baralho(self):
        """
        Começa um novo baralho com todas as cartas embaralhadas e esvazia a pilha de descartes
        """

        self._cartas = list(self._original)
        self.embaralhar()
        self.descarte.clear()
        self.em_jogo.clear()

    def embaralhar(self):
        """
        Embaralha as cartas disponíveis no baralho
        """
        novas_cartas = []

        while len(self._cartas) > 0:
            sacada = random.choice(self._cartas)
            novas_cartas.append(sacada)
            self._cartas.pop(self._cartas.index(sacada))

        self._cartas = novas_cartas

    # MÉTODOS DA PILHA ======================================================
    # Métodos que manipulam a pilha de cartas básica
    def ver_cartas(self, qtd: int) -> list:
        """
        Revela a quantidade selecionada de cartas do topo da pilha de saque
        :param qtd: a quantidade de cartas a serem vistas
        :return: lista com as cartas
        """
        return self._cartas[0:qtd]

    def sacar(self, do_topo: bool = True):
        """
        Saca uma carta do topo do baralho, removendo-a do mesmo
        :param do_topo: se marcado como falsa, pega aleatoriamente
        :return: a carta sacada, None se não tiver o que sacar
        """

        if len(self._cartas) > 0:
            sacada = self._cartas[0] if do_topo else random.choice(self._cartas)
            self.em_jogo.append(sacada)
            return self._cartas.pop(self._cartas.index(sacada))

        return None

    def descer_carta(self, carta, aleatorio: bool = False, embaralha: bool = False):
        """
        Coloca uma carta em jogo na pilha de saques
        :param carta: a carta a ser devolvida
        :param aleatorio: devolve aleatoriamente para a pilha, mas nunca como primeira carta. Se Falso, devolve para o final
        :param embaralha: reembaralha a pilha de saques ao devolver a carta
        """

        index = random.randint(1, len(self._cartas)) if aleatorio else len(self._cartas)
        self._cartas.insert(index, carta)
        self.em_jogo.pop(self.em_jogo.index(carta))

        if embaralha:
            self.embaralhar()

    def mover_cartas(self, qtd: int = 1, aleatorio: bool = True):
        """
        Move uma quantidade de cartas do início da pilha
        :param qtd: A quantidade de cartas a serem movidas
        :param aleatorio: Coloca aleatoriamente no baralho, se falso, coloca no final
        """
        a_mover = list(self._cartas[0:qtd])
        self._cartas = self._cartas[qtd:len(self._cartas)]

        for carta in a_mover:
            index = random.randint(1, len(self._cartas)) if aleatorio else len(self._cartas)
            self._cartas.insert(index, carta)

    # MÉTODOS DO DESCARTE ====================================================
    # Métodos que manipulam a pilha de descarte
    def descartar(self, carta):
        """
        Joga a carta selecionada na pilha de descartes
        :param carta: a carta a ser descartada
        """

        self.descarte.append(carta)
        self.descarte = list(reversed(self.descarte))
        self.em_jogo.pop(self.em_jogo.index(carta))

    def pegar_do_descarte(self, carta=None):
        """
        Pega uma carta da pilha de descartes. Se não escolher a carta, pega uma aleatória
        :param carta: qual a carta escolhida do descarte
        :return: a carta sacada, None se não tiver o que sacar do descarte
        """

        if len(self.descarte) > 0:
            sacada = carta if carta else random.choice(self.descarte)
            self.em_jogo.append(sacada)
            return self.descarte.pop(self.descarte.index(sacada))

        return None

    def ver_descarte(self, qtd: int = None) -> list | None:
        """
        Revela a quantidade selecionada de cartas do topo da pilha de descarte
        :param qtd: a quantidade de cartas a exibir, se omitido exibe a lista inteira
        :return: lista com as cartas
        """
        if not self.descarte:
            return None

        if not qtd:
            return self.descarte

        return self.descarte[0:qtd]

    # MÉTODOS DA PILHA E DO DESCARTE ========================================
    # Métodos que envolvem tanto a pilha de cartas e a pilha de descartes
    def recuperar_cartas(self, cartas_descartadas: list, aleatorio: bool = False, embaralha: bool = True):
        """
        Coloca as cartas selecionadas da pilha de descartes de volta para a pilha de saque
        :param cartas_descartadas: uma lista com as cartas a serem devolvidas da pilha de descarte
        :param aleatorio: se coloca aleatoriamente (True) ou no final da pilha (False - Padrão)
        :param embaralha: se embaralha a pilha de saques depois disso
        """
        while cartas_descartadas:
            sacada = self.pegar_do_descarte(cartas_descartadas[0])
            self.descer_carta(sacada, aleatorio, embaralha)


if __name__ == "__main__":
    baralho_teste = Baralho([carta for carta in range(1, 11)])
    print(baralho_teste)
    mao_extra = [baralho_teste.sacar() for _ in range(5)]
    print("Mão com 5 cartas_aplicadas:", mao_extra)
    print(baralho_teste)
    print("---------------------------------------------------------------------")
    print("Descartando as duas menores cartas da mão:")

    for i in range(2):
        sacado = min(mao_extra)
        baralho_teste.descartar(sacado)
        mao_extra.pop(mao_extra.index(sacado))

    print("Mão atual:", mao_extra)
    print("Pilha de descartes", baralho_teste.descarte)
    print(baralho_teste)



from .dado import Dado


# ==================================================================
# Funções para rolamento de _dados padrão, mas já retornando o total
# ==================================================================
def d4(qtd: int = 1) -> int:
    """
    Rola uma quantidade de **d4** e retorna a soma deles
    :param qtd: a quantidade de dados, 1 por padrão
    :return: valor total do rolamento
    """
    return Dado(qtd, 4).total


def d6(qtd: int = 1) -> int:
    """
    Rola uma quantidade de **d6** e retorna a soma deles
    :param qtd: a quantidade de dados, 1 por padrão
    :return: valor total do rolamento
    """
    return Dado(qtd, 6).total


def d8(qtd: int = 1) -> int:
    """
    Rola uma quantidade de **d8** e retorna a soma deles
    :param qtd: a quantidade de dados, 1 por padrão
    :return: valor total do rolamento
    """
    return Dado(qtd, 8).total


def d10(qtd: int = 1) -> int:
    """
    Rola uma quantidade de **d10** e retorna a soma deles
    :param qtd: a quantidade de dados, 1 por padrão
    :return: valor total do rolamento
    """
    return Dado(qtd, 10).total


def d12(qtd: int = 1) -> int:
    """
    Rola uma quantidade de **d12** e retorna a soma deles
    :param qtd: a quantidade de dados, 1 por padrão
    :return: valor total do rolamento
    """
    return Dado(qtd, 12).total


def d20(qtd: int = 1) -> int:
    """
    Rola uma quantidade de **d20** e retorna a soma deles
    :param qtd: a quantidade de dados, 1 por padrão
    :return: valor total do rolamento
    """
    return Dado(qtd, 20).total


def d66():
    """
    Rola um d66 e retorna o resultado somado. O d66 é gerado jogando dois d6 separadamente, um para a dezena e outro
    para a unidade
    :return: valor total do rolamento
    """
    dezena = d6() * 10
    unidade = d6()
    total = dezena + unidade

    return total


def d100(qtd: int = 1) -> int:
    """
    Rola uma quantidade de **d100** e retorna a soma deles
    :param qtd: a quantidade de dados, 1 por padrão
    :return: valor total do rolamento
    """
    return Dado(qtd, 100).total



from .helpers.ajeita_tabela import ajeita_tabela
from .helpers.ajusta_resultados import verifica_resultado_fora_alcance, rola_dados_resultado


def rolar_tabela(valor: int, tabela: list, coluna=None, rolar_dados=True) -> str:
    """
    Busca o resultado em uma tabela de RPG retornando a linha e a eventual
    coluna, caso tenha. Se o texto possuir notação de dados (exemplo: 1d6)
    também já rola o resultado.

    O padrão de cada linha é ((inicio, fim), resultado da busca).
    É possível colocar apenas um valor na primeira tupla caso esse item só tenha
    um valor. Resultado da busca pode ser uma string como "1d6 gp" ou um
    dicionário com o formato { "col 1": "resultado col 1", "col2": "valor col2" }

    :param valor: o valor a ser buscado
    :param tabela: a lista a ser usada conforme indicado acima.
    :param coluna: qual a coluna a ser buscada pelo resultado, se não tiver colunas, ignore
    :param rolar_dados: booleano para indicar se o resultado que contenha dados a ser rolado (ex.: "1d10 gp"). Se não quiser que já retorne o dado rolado,     marque False

    :return:
    """
    # Ajeita a _tabela para o caso da tupla não ter dois itens
    nova_tabela = ajeita_tabela(tabela)

    # pega os índices a serem buscados
    indice_menor = nova_tabela[0][0][0]
    indice_maior = nova_tabela[-1][0][1]

    # Busca a linha da _tabela
    resultado = next(
        (item for (x, y), item in nova_tabela if x <= valor <= y),
        None
    )

    resultado = verifica_resultado_fora_alcance(valor, indice_menor, indice_maior, nova_tabela, resultado)

    # se tiver mais de uma coluna, seleciona a coluna
    resultado = resultado[coluna] if coluna else resultado

    # transforma o texto que contenha notação de _dados em texto com resultado
    resultado = rola_dados_resultado(rolar_dados, resultado)

    return resultado


if __name__ == "__main__":
    # Exemplo de uso
    exemplo_tabela = [
        ((1, 6), {"coluna 1": "2d6 moedas de ouro", "coluna 2": "2d10+1d6 moedas de ouro e 1d6 de prata",
                  "coluna 3": "d8*2 moedas de ouro"}),
        ((7, 9), {"coluna 1": "2d6 moedas de ouro", "coluna 2": "2d10+1d6 moedas de ouro e 1d6 de prata",
                  "coluna 3": "d8*2 moedas de ouro"}),
        (10, {"coluna 1": "2d6 moedas de ouro", "coluna 2": "2d10+1d6+100 moedas de ouro e 1d6 de prata",
              "coluna 3": "d8*2 moedas de ouro"})
    ]

    print('-------------------------------------------------------------------------')
    print("Tabela:")
    print(*exemplo_tabela, sep='\n')
    print('-------------------------------------------------------------------------')
    print(f"O resultado foi: {rolar_tabela(rolar_dado_notacao('d10'), exemplo_tabela, 'coluna 2')}")
    print('-------------------------------------------------------------------------')


import re
from helpers.monta_dado import monta_dado


def converter_notacao_dado(notacao: str) -> dict:
    """
    Converte a notação de _dados de RPG em um objeto para ser manipulado posteriormente

    :param notacao: a notação de dado usada. Exemplo: "3d6+1d4+3"
    :return: um dicionário: { "_dados": [...dado], "modificador_dado": { "modificador_dado": numero, "operador": string }
        "dado": { "quantidade": numero, "faces": numero, "operador": string }
        No exemplo acima: {
            "_dados": [ { "quantidade": 3, "faces": 6 }, { "quantidade": 1, "faces": 4", "operador": "+" } ],
            "modificador_dado": { "numero": 3, "operador": "+" }
        }
    """

    # Separa a fórmula em partes
    partes = re.split(r"([+\-*/])", notacao)

    # Separa o modificador_dado numérico
    if "d" in partes[-1]:
        modificador = {
            "valor": 0,
            "operador": "+"
        }
    else:
        modificador = {
            "valor": int(partes.pop(-1)),
            "operador": partes.pop(-1)
        }

    # Converte o primeiro dado na variável
    dados = [monta_dado(partes.pop(0))]

    # Insere as demais partes
    while partes:
        operador = partes.pop(0)
        print(type(operador))
        notacao_dado = partes.pop(0)

        dado = monta_dado(notacao_dado)
        dado["operador"] = operador

        dados.append(dado)

    resultado = {
        "_dados": dados,
        "modificador_dado": modificador
    }

    return resultado



def monta_dado(string_dado: str) -> dict:
    separando_elementos = string_dado.lower().split("d")
    separando_elementos[0] = separando_elementos[0] if separando_elementos[0] else '1'

    quantidade, faces = map(int, separando_elementos)

    return {"quantidade": quantidade, "faces": faces}


def ajeita_tabela(tabela: list[tuple, str]) -> list[tuple, str]:
    """
    Pega uma _tabela que tenha itens com apenas um rolamento e coloca na tupla como valor mínimo e máximo
    :param tabela: a _tabela a ser arrumada
    :return: lista
    """
    nova_tabela = []

    for item in tabela:
        alcance_rolamento = item[0]
        resultado_rolamento = item[1]

        if type(alcance_rolamento) is not tuple:
            alcance_rolamento = (int(alcance_rolamento), int(alcance_rolamento))
        else:
            alcance_rolamento = (int(alcance_rolamento[0]), alcance_rolamento[1])

        linha = (alcance_rolamento, resultado_rolamento)
        nova_tabela.append(linha)


    return nova_tabela



from RPG.tabela_rolavel import TabelaRolavel
from .data.criticos import criticos

tabela_criticos = {chave: TabelaRolavel(criticos[chave], 'd100') for chave in criticos.keys()}
intensidades = 'ABCDE'


def consulta_critico(tipo_critico: str, intensidade_critico: str, valor_rolado: int) -> str:
    """
    Faz uma consulta na tabela de críticos com um valor rodado externamente e retornar o resultado
    :param tipo_critico: qual o crítico procurado, como chave da tabela de críticos
    :param intensidade_critico: qual a letra do crítico
    :param valor_rolado: qual o valor rolado nos dados
    :return: uma string com o resultado do crítico
    """
    intensidade_critico = intensidade_critico.upper()
    valor_rolado = 100 if valor_rolado == 0 else valor_rolado

    if tipo_critico not in criticos.keys():
        print(f'{tipo_critico} não é um tipo de crítico válido. Tipos aceitáveis: {", ".join(list(criticos.keys()))}.')
        return ''

    if intensidade_critico not in intensidades:
        print(f'{intensidade_critico} não é uma intensidade de crítico válida. Valores aceitáveis: {", ".join(list(intensidades))}')
        return ''

    critico = tabela_criticos[tipo_critico]
    critico.rolar(valor_fixo=valor_rolado, coluna=intensidade_critico)

    return critico.resultado



# BÁSICO
from .dado import *
from .tabela_rolavel import rolar_tabela, TabelaRolavel
from .baralho import Baralho
from .iniciativa import ordena_iniciativa

# SISTEMAS
from .sistemas import dnd
from .sistemas import rolemaster


from RPG.dado import d20
from helpers.dnd.checa_testes import checa_teste, checa_ataque


class Teste:

    def __init__(self, modificador=0, dc=10, vantagem=False, desvantagem=False):
        """
        Define o teste para uma habilidade
        :param modificador: o modificador no rolamento
        :param dc: qual o DC do teste
        :param vantagem: se o personagem tem vantagem na rolagem
        :param desvantagem: se o personagem tem desvantagem na rolagem
        """
        self._modificador: int = modificador
        self._dc: int = dc
        self._vantagem: bool = vantagem
        self._desvantagem: bool = desvantagem
        self._rolamento = []

        self._corrige_vantagem_desvantagem()
        self.dado = self._rolamento_inicial()

    def __str__(self):
        string_dado = 'dado rolado' if len(self._rolamento) == 1 else 'dados rolados'
        string_modificador = self._define_str_modificador()

        return f'{string_dado.capitalize()}: {", ".join(str(d) for d in self._rolamento)}' \
               f'. Modificador: {string_modificador}{self._define_str_vantagem_desvantagem()}.' \
               f' Total: {self.total} vs DC {self._dc}.' \
               f' {self.resultado.capitalize()}.'

    def __repr__(self):
        return f'<Teste{self._define_str_vantagem_desvantagem()},' \
               f' dc: {self._dc}, rolamento={self._rolamento},' \
               f' modificador: {self._define_str_modificador()}>'

    # INICIALIZADORES E AUXILIARES=====================================================
    def _rolamento_inicial(self):
        self._rolamento.append(d20())

        if self._vantagem or self._desvantagem:
            self._rolamento.append(d20())
            return max(self._rolamento) if self._vantagem else min(self._rolamento)

        return self._rolamento[0]

    def _corrige_vantagem_desvantagem(self):
        if self._vantagem and self._desvantagem:
            self._vantagem = False
            self._desvantagem = False

    def _define_str_vantagem_desvantagem(self) -> str:
        return '' if not (self._vantagem or self._desvantagem) else \
            ' com vantagem' if self._vantagem else ' com desvantagem'

    def _define_str_modificador(self) -> str:
        return f'+{self._modificador}' if self._modificador > - 1 else self._modificador

    # MÉTODOS =========================================================================
    def rerolar(self, modificador=None, dc=None, vantagem=None, desvantagem=None) -> None:
        self._modificador = modificador if modificador else self._modificador
        self._dc = dc if dc else self._dc
        self._vantagem = vantagem if vantagem else self._vantagem
        self._desvantagem = desvantagem if desvantagem else self._desvantagem

        self._rolamento.clear()
        self.dado = self._rolamento_inicial()

    # PROPRIEDADES ====================================================================
    @property
    def total(self) -> int:
        dado = self.dado
        resultado = dado + self._modificador

        return resultado

    @property
    def resultado(self) -> str:
        return checa_teste(self.total, self._dc)

    @property
    def dado_minimo(self) -> int:
        """
        Calcula o valor mínimo que precisa ser tirado no d20 para conseguir um sucesso
        :return: Um valor entre 1 a 20. Se o resultado for None, não é possível passar nesse teste
        """
        rolamento_min = self._dc - self._modificador
        return rolamento_min if rolamento_min in range(1, 21) else 1 if rolamento_min <= 0 else None

    @property
    def chance_sucesso(self) -> str:
        dado = self.dado_minimo if self.dado_minimo else 0
        return str((21 - dado) * 5) + '%'


class Ataque(Teste):
    def __init__(self, modificador=0, ac=10, vantagem=None, desvantagem=None):
        super().__init__(modificador, dc=ac, vantagem=vantagem, desvantagem=desvantagem)
        self._ac = self._dc

    def __str__(self):
        string_dado = 'dado rolado' if len(self._rolamento) == 1 else 'dados rolados'
        string_modificador = self._define_str_modificador()

        return f'{string_dado.capitalize()}: {", ".join(str(d) for d in self._rolamento)}' \
               f'. Modificador: {string_modificador}{self._define_str_vantagem_desvantagem()}.' \
               f' Total: {self.total} vs AC {self._ac}.' \
               f' {self.resultado.capitalize()}.'

    def __repr__(self):
        return f'<Ataque{self._define_str_vantagem_desvantagem()},' \
               f' ac: {self._ac}, rolamento={self._rolamento},' \
               f' modificador: {self._define_str_modificador()}>'

    # PROPRIEDADES ====================================================================
    @property
    def resultado(self) -> str:
        return checa_ataque(self.dado, self.total, self._ac)

    @property
    def dado_minimo(self) -> int:
        """
        Calcula o valor mínimo que precisa ser tirado no d20 para conseguir um sucesso
        :return: Um valor entre 1 a 20. Se o resultado for None, não é possível passar nesse teste
        """
        rolamento_min = self._dc - self._modificador
        return rolamento_min if rolamento_min in range(1, 21) else 1 if rolamento_min <= 0 else 20



resultado_rolagem_teste = ('Falha', 'Sucesso')
resultado_rolagem_ataque = ('Erro', 'Acerto', 'Crítico')

tipos_moedas = ['cp', 'sp', 'ep', 'gp', 'pp']

valor_moedas = {
    'cp': 1/100,
    'sp': 1/10,
    'ep': 1/2,
    'gp': 1,
    'pp': 10
}


from .rolar_tabela import rolar_tabela
from .helpers.ajeita_tabela import ajeita_tabela
from .helpers.busca_index_item_tabela import busca_index_item_tabela
from .helpers.rolamento_dado import DadoRolado
from RPG.dado import rolar_dado_notacao


class TabelaRolavel:
    def __init__(self, tabela: list = None, dados: str | int = None):
        """
        Cria um objeto com uma tabela vazia ou já existente para ser rolada e métodos para que busquem o
        resultado diretamente nela.

        :param tabela: a tabela a ser buscada, se não passar o parâmetro, cria uma tabela vazia (que deve ser preenchida com .insere_linha_tabela()
        :param dados: a notação de dados a ser usado para rolar na tabela. Também pode ser passado um valor fixo (que pode ser alterado com o método .rolar())
        """
        self._tabela = ajeita_tabela(tabela) if tabela else []
        self._dados = dados
        self._dado_rolado = DadoRolado(rolamento=0, modificador=0)
        self._resultado = ''

        if dados:
            self.rolar()

    def __repr__(self):
        return f'<Tabela Rolável={self._tabela},\n dado usado: {self._dados}>'

    def __str__(self):
        return f'Rolamento de {self._dados} + mod ({self.rolamento.modificador}): {self.rolamento.total}. Resultado: {self._resultado}'

    # Métodos de edição da _tabela ====================================================================================
    def insere_linha_tabela(self, valor_menor: int, valor_maior: int, resultado) -> None:
        self._tabela.append([(valor_menor, valor_maior), resultado])

    def edita_alcance_da_linha(self, valor_da_busca: int, novo_minimo: int, novo_maximo: int) -> None:
        """
        Altera o alcance de uma linha da _tabela.

        :param valor_da_busca: Um valor que esteja entre o alcance da _tabela anterior
        (exemplo, se quiser alterar (1,10) para (1, 15) o valor_da_busca pode ser qualquer número entre 1 e 10
        :param novo_minimo: O novo valor mínimo
        :param novo_maximo: O novo valor máximo
        """
        nova_tupla = (novo_minimo, novo_maximo)
        index = busca_index_item_tabela(self._tabela, valor_da_busca)

        self._tabela[index][0] = nova_tupla

    def edita_resultado_da_linha(self, valor_da_busca: int, novo_resultado) -> None:
        """
        Altera o resultado que pode ser obtido naquela linha. Exemplo: [(1, 10), **"resultado 1"**] pode virar [(1,
        10), **"tesouro 1"**]

        :param valor_da_busca: Um valor que esteja entre o alcance da _tabela anterior (exemplo,
        se quiser alterar (1,10) para (1, 15) o valor_da_busca pode ser qualquer número entre 1 e 10
        :param novo_resultado: O texto a ser substituído
        """
        index = busca_index_item_tabela(self._tabela, valor_da_busca)
        self._tabela[index][1] = novo_resultado

    # Métodos de resultado da rolagem ================================================================================
    def rolar(self, valor_fixo: int = None, modificador_dado: int = 0, coluna: str = None) -> None:
        """
        Rola novamente a tabela e altera o resultado. Tem a opção de incluir a coluna escolhida e com isso rolar
        eventuais dados do texto (exemplo: "3d6 moedas"). Se preferir o resultado puro de texto, não passe a coluna
        como parâmetro, mas use objeto.resultado["coluna"] no lugar
        :param valor_fixo: Procura por um valor específico dentro da tabela. Se omitido vai pegar o valor passado na atribuição da classe em "dados"
        :param modificador_dado: Para o caso de haver algum modificador não básico aplicado na tabela (ex.: role na tabela X com +10)
        :param coluna: Caso o resultado tenha mais de uma coluna. Seleciona a coluna e rola dos dados dentro dela
        """
        valor_fixo = self._dados if not valor_fixo else valor_fixo
        dado = valor_fixo if type(valor_fixo) == int else rolar_dado_notacao(valor_fixo)

        self._dado_rolado = DadoRolado(dado, modificador_dado)
        self._resultado = rolar_tabela(self._dado_rolado.total, self._tabela, coluna)

    # Getters ========================================================================================================
    @property
    def tabela(self):
        return self._tabela

    @property
    def rolamento(self):
        return self._dado_rolado

    @property
    def resultado(self):
        return self._resultado
